/*********************************************************/
/*  This source code is copyrighted by the author        */
/*  Please read COPYING  for the GNU licence policy.     */
/*********************************************************/
#include <vga.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include "tetX.h"
#include "egafont.h"

#ifdef HAVE_DIRENT_H 
#include <dirent.h>
#endif

#ifndef P_tmpdir
#define P_tmpdir "/tmp"
#endif

static char *Font;
static int  FontHeight;
static int  charset_size;
int	wide_chars;


int egaShowText(char *text, int x, int y, int fgcol, int bgcol)
{

   unsigned char buff[160];		// Enough for 20 char-s
   unsigned char *buffer = NULL;
   unsigned char mask;
   int  buflen, i;
   unsigned char fbyte;
   unsigned char *to;
   int      len,  rc;
   tchar  *textbuf, *textptr, textchar;     

   rc = 4;
   
#ifdef TET_USE_WIDE_CHARS
   buflen = 0;	
   len = mbstowcs(NULL, text, 0);
   textbuf = malloc((len+1) * sizeof(wchar_t));
   if (textbuf == NULL) goto HastaLaVista;
   mbstowcs(textbuf, text, len+1);
#else   
   textbuf = text;
   len  = strlen(text);
#endif
   
   buflen = len << 3;
   // Attempting to save time for short texts
   
   if (buflen > 160)
   {  buffer = (char *) malloc(buflen);
      if (!buffer) goto HastaLaVista;
   } 
   else  buffer = buff;

   for (i=0; i<FontHeight; i++)
   { 
      textptr = textbuf;
      to = buffer;
	  
      while ((textchar = *textptr++) != (tchar)0)
      {  fbyte = *(Font + textchar*FontHeight + i);
         mask = 0x80;
         while(mask)
         {  *to++ = (fbyte & mask) ? fgcol: bgcol;
            mask = mask >> 1;
         }
      }
    
      setrow(x, y+i, buflen, buffer);
   }
   
   rc = 0;

HastaLaVista:
   if (buflen > 160 && buffer) free (buffer);

#ifdef TET_USE_WIDE_CHARS
   if (textbuf) free(textbuf); 
#endif

   return rc;

}

int egaShowScaledText(char *text, int x, int y,
                      int fgcol, int bgcol, int hscale, int vscale)
{
   unsigned char *buffer;
   unsigned char mask;
   int  len,buflen,i, j;
   unsigned char fbyte, c;
   unsigned char *to;
   tchar  *textbuf, *textptr, textchar;
   int rc = 4;
   
#ifdef TET_USE_WIDE_CHARS
   buffer = NULL;	
   len = mbstowcs(NULL, text, 0);
   textbuf = malloc((len+1) * sizeof(wchar_t));
   if (textbuf == NULL) goto HastaLaVista;
   mbstowcs(textbuf, text, len+1);
#else   
   textbuf = text;
   len  = strlen(text);
#endif

   buflen = (len << 3) * hscale;
   
   //  This is called rarely, so you don't have to worry about time 
   buffer = (char *) malloc(buflen);
   if (buffer == NULL) goto HastaLaVista;

   for (i=0; i<FontHeight; i++)
   {
      textptr = textbuf;
      to = buffer;

      while ((textchar = *textptr++) != (tchar)0)
      {  fbyte = *(Font + textchar*FontHeight + i);
         mask = 0x80;
 
         while(mask)
         {  c =  (fbyte & mask) ? fgcol: bgcol;
            j = hscale;
            while(--j >=0) *to++ = c;
            mask = mask >> 1;
         }
      }
      
      j = vscale;
      while (--j >= 0)
         setrow(x, y++, buflen, buffer);
   }

   rc = 0;
  
HastaLaVista:

  if (buffer) free (buffer);

#ifdef TET_USE_WIDE_CHARS
   if (textbuf) free(textbuf); 
#endif

   return rc;

}

int egaSetFont(char *cFont)
{   int *iFont;
    if (!cFont) return 4;

    iFont = (int *)cFont;
    FontHeight = *iFont;
    
    Font = cFont + sizeof(int);
    return 0;
}
  
void egaFreeFont(char *cFont)
{
   if (!cFont) return;

   if (cFont + sizeof(int) == Font)
   {  Font = NULL; FontHeight = 0; }

   free(cFont);
}        


int egaGetFontHeight(void)
{   return FontHeight; }

int egaGetTextWidth(char *text)
{
   int len;	
#ifdef TET_USE_WIDE_CHARS
    len = mbstowcs(NULL, text, 0);
#else   
    len  = strlen(text);
#endif
	return len * 8;
}

char *egaGetFontPtr(int chr)
{  return Font + (chr & 0xff) *FontHeight; }


char *egaLoadFont(char *fontFileName, char *fontDirName, char *mes)
{
   int fd = -1; 
 
   int rc = 4; 
   size_t s;
   off_t  hdrlen;    // File header length

   char  *cFont = NULL;
   int   *iFont;
   char *FontPath=NULL, *FontPath1=NULL, *tmp_file_name=NULL;
   unsigned int fontheight;
   static const char *extensions[] = {"", ".gz", ".psf", ".psf.gz", NULL};	
   const char *extension;	
   int   unzipped;
   static const char *tmp_file_pref = "itet";
   unsigned char  strt[4];
  

#ifdef HAVE_DIRENT_H
   if (memcmp(fontDirName, "/usr", 4) == 0)
   {
	DIR *DirTest = opendir(fontDirName);
	
	if (DirTest == NULL)
	{	
		DirTest = opendir(fontDirName+4);
		if (DirTest) strcpy(fontDirName, fontDirName+4);
	}	

	if (DirTest == NULL) return NULL;
	closedir(DirTest);
   }
#endif	


   wide_chars = 0;	   
   FontPath = malloc(strlen(fontDirName) + strlen(fontFileName) + 5);
   strcpy(FontPath, fontDirName);
   fontheight = strlen(FontPath);
   if (FontPath[fontheight-1] != '/') strcat(FontPath, "/");
   strcat(FontPath, fontFileName);
   FontPath1 = GetFullFileName(FontPath);    
   free(FontPath);

	

   FontPath = malloc(strlen(FontPath1) + 11);
   fontheight = strlen(FontPath1);
   memcpy(FontPath, FontPath1, fontheight); 
   free(FontPath1); FontPath1 = NULL;
		
   for (unzipped=0; (extension=extensions[unzipped])!=NULL; unzipped++)
   { 		
	strcpy(FontPath + fontheight, extension);
	fd = open(FontPath, O_RDONLY);
	if (fd >= 0) break;
   }

   if (fd < 0)
   {  FontPath[fontheight] = '\0';
      sprintf(mes, "Could not open font %s\n", FontPath);
      free(FontPath);
	  return NULL;
   }		

   fontheight += strlen(extension);
   unzipped = strcasecmp(FontPath+fontheight-3, ".gz");

   if (!unzipped)
   {
	const char  *tmp_dir_name;

	close(fd);  fd = -1;
	tmp_dir_name = getenv("TMPDIR");
	if (tmp_dir_name == NULL) tmp_dir_name = P_tmpdir;

	FontPath1 = malloc(strlen(FontPath) + strlen(tmp_dir_name)
                        + strlen(tmp_file_pref) + 50);
	if (FontPath1 == NULL) goto Ret;

 	 sprintf(FontPath1, "gunzip -c %s > ", FontPath);
	 tmp_file_name = FontPath1 + strlen(FontPath1);
	 sprintf(tmp_file_name, "%s/%s%08x.cfont",
		tmp_dir_name, tmp_file_pref, getpid());
         system(FontPath1);

         fd = open(tmp_file_name, O_RDONLY);
   }       

   free(FontPath);
   
   if (fd < 0) return NULL;

   /* Try to find the size of characters */
   read(fd, strt, 4);
   charset_size = 256;
   
	/* Removed check for bytes pers char,
       because it will work with wide chars as well */ 
   if (strt[0] == 0x36 && strt[1] == 4) /* && strt[2] <= 1) */
   {  /* psf file */
       fontheight = strt[3];
#ifdef TET_USE_WIDE_CHARS 
	s = lseek(fd, 0, SEEK_END) - 4;
       lseek(fd, 4, SEEK_SET); 
	wide_chars = 1;
#else
       s = fontheight<<8;
#endif	   
   }
   else
   {   /* Size of the font file */
       s = (size_t) lseek(fd, 0, SEEK_END);

       fontheight = s>>8;       // Assume 256 characters, FontHeight bytes each
       hdrlen  = s & 255;
       if (hdrlen != 0 && hdrlen != 4) goto Ret;
       s -= hdrlen;

       lseek(fd, hdrlen, SEEK_SET);  // Bypassing header
   }

   cFont = (char *) malloc(s + sizeof(int));
   if (!cFont) goto Ret;

   iFont = (int *)cFont;
   *iFont = fontheight;

   s = read(fd, cFont+sizeof(int), s);
   if (s > 0) 
	{	rc = 0; 
   		charset_size = s/fontheight;
	}	
 
Ret:  
   if (fd >= 0) close(fd);

   if (!unzipped)	
   {
	if(tmp_file_name) remove(tmp_file_name);
	free(FontPath1);		
   }	
	
   if (rc)
   {  if (cFont) free(cFont);
      return NULL;
   }

   return cFont;
}

int egaShowCharSpecial(tchar b, int x, int y, int fgcol, int bgcol,
                       int hscale, int vscale, int flags)
{   char *buffer, *bufptr, c;
    unsigned char *fontptr, fbyte, mask, smask;
    int hflip;
    int i, y1, yincr, w;
    
    w = hscale << 3; 
    buffer = (char *) malloc(w);
    if (!buffer) return 4;
 
    hflip = flags & EGAFONT_HFLIP;
    if (flags & EGAFONT_VFLIP)
    { y += vscale * FontHeight - 1;
      yincr = -1;
    }
    else
      yincr = 1;
    
    smask = hflip? 1 : 0x80;
    
    b &= 0xff;
    fontptr = Font + b*FontHeight;
 
    y1 = FontHeight;
    while(--y1 >= 0)
    {  
        fbyte = *fontptr++;
        bufptr = buffer;
        mask = smask;
		while (mask)  
        {  c = fbyte & mask ? fgcol : bgcol;
           for ( i=0; i<hscale; i++) *bufptr++ = c;
           if (hflip) mask <<=1; else mask >>= 1; 
        }   

        for (i=0; i<vscale; i++) 
        {   setrow(x, y, w, buffer);
            y += yincr;
        }
     }

     free(buffer);
     return 0; 
} 
