/*===================================================*/
/*          Joystick and keyboard processing         */
/*---------------------------------------------------*/
/*       04.04.98                          M.G.      */
/*===================================================*/
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <ctype.h>
#include "tetX.h"


#ifdef X11

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#ifdef JOYSTICK_XINPUT
#include <X11/extensions/XInput.h>
#endif  /* def JOYSTICK_XINPUT */
#else	/* def X11 */	
#include <vga.h>
#undef JOYSTICK_XINPUT
#endif	/* def X11 */	



#include "kbcodes.h"
#define Esc 0x1B


#ifdef JOYSTICK_SUPPORTED
static int JoyX1=0, JoyX2=0, JoyY1=0, JoyY2=0;
static int JoyX=0, JoyY=0, Buttons=0;
/* static int JoyButtonLocked; */
char JoyName[32] = "unknown";
int  JoyButtonMask = 0;  /* Button auto-repeat */
/*
#warning  "If joystick support is not included in your"
#warning  "system, you need to comment out line "
#warning  "'#define JOYSTICK_SUPPORTED' in tetconfig.h,"
#warning  "otherwise this file will compile with errors"
*/

#ifdef  JOYSTICK_DRIVER
#include <linux/joystick.h>
#include <sys/ioctl.h>
#ifdef JS_VERSION  
#if (JS_VERSION >= 0x010208) && (JOYSTICK_DRIVER == 1)
#define NEW_JOYSTICK 
#define NEW_JOY_MAX 32767l
#endif   /* JS_VERSION > 0x010208 */
#endif   /* def JS_VERSION */
#endif   /* def  JOYSTICK_DRIVER */


#else    /* JOYSTICK_SUPPORTED */
char JoyName[] = "not supported";
#endif   /* JOYSTICK_SUPPORTED */

extern int  JoyEnabled;                // tetgame.c
static int JoyFd = -1;
short  UseJoystick = 1;    /* 0 - old driver interface,
                              1 - new driver interface
                              2 - XInput */
short  JoyNewDriver = 0;
int    JoyVersion = 0;


/* Get a key from the keyboard if available
   For special keys returns
        LSB:  ASCI code or last byte in ANSI sequence
        MSB:  0 - ASCII character
	        1 - arrow key
              2 - Alt key
	        3 - Functional  key (F1, F2, etc) with no CTRL
*/

#ifdef X11
extern  Display		*display;
extern  Window		window;
extern  Window		root_window;

int ProceedNonkeyEvent(XEvent *event);

#ifdef JOYSTICK_XINPUT
static  XDevice *xdev = NULL;
static  int button_press_type = -1;
static  int button_release_type = -1;
static  int motion_notify_type = -1;
static  int JoyCX=0, JoyCY=0;	/* Joystick middle position */
static  int OldJoyX = 0, OldJoyY = 0; 
#endif	/* def JOYSTICK_XINPUT */


int ProcessKeyEvent(XKeyEvent *key_event)
{
      char str[3]; 
      int  n;
      KeySym keysym;

      
      n = XLookupString(key_event, str, 2, &keysym, (XComposeStatus *)NULL);    

           
      if (keysym == XK_KP_Begin || keysym == XK_Begin
                    || keysym == 0) keysym = XK_Execute;             

      if (keysym >= XK_KP_0 && keysym <= XK_KP_9)
                    keysym = (keysym - XK_KP_0) + '0'; 

      if (keysym >= XK_KP_Home && keysym <= XK_KP_End)
                    keysym = (keysym - XK_KP_Home) + XK_Home;
		    		          
      
      return (int) keysym;
}

unsigned int  kbdGet(void)
{
  int EventCounter;
  XEvent evnt;

  if (GetWindowState() == IconicState) SuspendGame();
  EventCounter = XEventsQueued(display, QueuedAlready);
//  EventCounter = XPending(display);

  while (--EventCounter >= 0)
  {
	XNextEvent(display, &evnt);
	if (XFilterEvent(&evnt, window)) continue;
	
	if (evnt.type == KeyPress) 
	   return ProcessKeyEvent((XKeyEvent *)&evnt);

	ProceedNonkeyEvent(&evnt); 
	   
  }
  
  return 0;
}


#else    /* def X11 */
int  kbdGet(void)
{
   int key;
   int buffer[10];
   int *buf;

   key = vga_getkey();
   if (key == KY_ERASE) key = KY_BACK;
   if (key != Esc) return key;

   // Key == Esc look for special key
   buf = buffer;
   while((*buf++=vga_getkey()) != 0);

   buf = buffer;
   if ((key=*buf++) != 0x5b) return key ? 0x200 | key : Esc;    // Meta key
   if ((key=*buf++) != 0x5b) return 0x100 | key;    // Arrow key
   return 0x300 | *buf;			   // Functional key

}
#endif	/* def X11 */



/* Init joystick   */
/* Returns
    0 - all right
    4 - error
*/

#ifdef JOYSTICK_XINPUT
Status PrepareXInputJoystick(void)
{	
    XEventClass xec[3];
    XDeviceInfo *xdi, *xdi1;
    XExtensionVersion *xev = NULL;    
    XID  devid;

    char **extensions, **extension_name = NULL;
    char *devname = NULL;
    int ndevs;

    JoyVersion = 0x00000;	

    extension_name = extensions = XListExtensions (display, &ndevs);
    if (extensions == NULL) return False;
    
    while(--ndevs>=0 &&
          strcmp(*extension_name++, "XInputExtension"));

    XFreeExtensionList(extensions);	  
    if (ndevs < 0) return False;    

    
    xev = XGetExtensionVersion(display, "JOYSTICK");
    if (xev)
    {    
       JoyVersion = ((xev->major_version)<<8) +
	            (xev->minor_version);        
       XFree(xev); xev = NULL;
    }

    devid = (XID) None;    
    xdi1 = XListInputDevices(display, &ndevs);
    if (xdi1 == NULL) return False;
    
    xdi = xdi1;
    while (--ndevs >= 0)
    { if (xdi->use == IsXExtensionDevice)
      {
         devname = xdi->name;
	 if (strncasecmp(devname, "joy", 3) == 0 || 
             strncasecmp(devname, "jstk", 4) == 0)
         {
             snprintf (JoyName, 31, "Device name: %s", devname);
             JoyName[32] = '\0';
             devid = xdi->id;
	     break;
         }
      }
      
      xdi ++;
    }
   
    XFreeDeviceList(xdi1);  
    if (ndevs < 0) return False; 
 
    xdev = XOpenDevice(display, devid);
    if (xdev == NULL) return False;

    DeviceButtonPress(xdev, button_press_type, xec[0]);
    DeviceButtonRelease(xdev, button_release_type, xec[1]);
    DeviceMotionNotify(xdev, motion_notify_type, xec[2]);
    
    XSelectExtensionEvent(display, root_window, xec, 3);

    return  True;

}
#endif  /* def JOYSTICK_XINPUT */


int joyInit(void)
{
#ifdef JOYSTICK_SUPPORTED

#ifdef JOYSTICK_DRIVER
#ifdef JOYSTICK_XINPUT
    if (UseJoystick == 2) goto Bypass;
#endif

    if (JoyFd >= 0) close(JoyFd);
    JoyFd = open ("/dev/js0", O_RDONLY | O_NONBLOCK);
    
    if (JoyFd >= 0)
    { JoyVersion = 0x000800;	
      JoyNewDriver = 0;

#ifdef NEW_JOYSTICK
       if (UseJoystick && JoyFd >=0 && ioctl(JoyFd, JSIOCGVERSION, &JoyVersion) >= 0)
       {   JoyNewDriver = (JoyVersion >= 0x010208);
           if (JoyNewDriver)
           {// fcntl(JoyFd, F_SETFL, O_NONBLOCK);
              ioctl(JoyFd, JSIOCGNAME(31), &JoyName); 
	      JoyName[31] = '\0';
           }    
       }
#endif	 /* def NEW_JOYSTICK */
       return 0;
    }      

#ifdef JOYSTICK_XINPUT
Bypass:
#endif
#endif	 /* def JOYSTICK_DRIVER */

#ifdef JOYSTICK_XINPUT
    if (PrepareXInputJoystick()) return 0;
#endif     /* def JOYSTICK_XINPUT */

#endif   /* def JOYSTICK_SUPPORTED */
    return 4;    
}

void joyDeinit(void)
{
#ifdef JOYSTICK_XINPUT
    if (xdev != NULL)
    {	XCloseDevice (display, xdev);
	xdev = NULL;
    }	
#endif    	
#ifdef JOYSTICK_DRIVER
     if (JoyFd >= 0) close (JoyFd);
     JoyFd = -1;
#endif
}

int joyGetStatus(void)
{

#ifdef JOYSTICK_SUPPORTED
     static int  OldButtons = 0;

#ifdef JOYSTICK_XINPUT

     if (xdev)
     {
/*        XDeviceState *xds, *xds1;
	XInputClass   *xis;		     
	XButtonState *xbs;
	XValuatorState *xvs;  */
//	int flag = 0;
	XDeviceMotionEvent m_event;
    XDeviceButtonEvent b_event;
	int NewButtons = 0;
	int NewJoyX, NewJoyY; 


	NewJoyX = JoyCX;  NewJoyY = JoyCY; 
    while (XCheckTypedEvent(display, motion_notify_type, (XEvent *) &m_event))
    {
	   NewJoyX = m_event.axis_data[0];
	   NewJoyY = m_event.axis_data[1];
	}
	
/*	if ( OldJoyX >= JoyX1 && OldJoyX <= JoyX2 &&
	     OldJoyY >= JoyY1 && OldJoyY <= JoyY2)  */
	if ( OldJoyX != NewJoyX || OldJoyY != NewJoyY) 
	{  JoyX = NewJoyX, JoyY = NewJoyY; }
	else
	{  JoyX = JoyCX; JoyY = JoyCY; }
	       
	OldJoyX = NewJoyX; OldJoyY = NewJoyY;
	
	NewButtons = Buttons;
	
       while(XCheckTypedEvent(display, button_press_type, (XEvent *) &b_event))
           NewButtons |= 1<<(b_event.button-1);
	   
       while(XCheckTypedEvent(display, button_release_type, (XEvent *) &b_event))
           NewButtons &=  ~(1<<(b_event.button-1));
/*
        xds = xds1 = XQueryDeviceState(display, xdev);
        if (xds == NULL) return 4;
	ndevs = xds->num_classes;
	xis = xds->data;

	flag = 0;

	while (--ndevs >= 0)
	{ if (xis->class == ButtonClass)
	  {  flag |= 1;
	     xbs = (XButtonState *)xis;
	     NewButtons = (xbs->buttons[1] << 7) |
	                  ( (xbs->buttons[0] >> 1) & 0x7f);	      
	  }		  
	  else    
	  if (xis->class == ValuatorClass)
	  {  flag |= 2;
	     xvs = (XValuatorState *)xis;
	     JoyX = xvs->valuators[0];
	     JoyY = xvs->valuators[1];
  	  }

	  if (flag == 3) break;	  
	  xis = (XInputClass *)((char *)xis + xis->length);
	}
	
        XFreeDeviceState(xds1);
	if (flag != 3) return 4;
*/	
	Buttons = NewButtons & ((NewButtons ^ OldButtons) | JoyButtonMask);
	OldButtons = NewButtons;
	return 0;
     }	

#endif   /* def JOYSTICK_XINPUT */

#ifdef JOYSTICK_DRIVER
/*    if (JoyFd < 0)
      {  joyInit();
         if (JoyFd < 0) return 4;
      }
*/
#ifdef NEW_JOYSTICK

     if (JoyNewDriver)
     {  struct js_event joyEvent;
        int  mask;
        __u8 eType, eNumber;
        __s16 eValue;
        char new_status = 0;

	Buttons &= JoyButtonMask;
        while(read(JoyFd, &joyEvent, sizeof(struct js_event)) > 0)
        {	
	   new_status = 1;
           eType = joyEvent.type; 
	   eNumber = joyEvent.number;

  	   eValue = joyEvent.value;

	   switch (eType & ~JS_EVENT_INIT)
	   { 	case JS_EVENT_BUTTON:
                   if ( eNumber < 4)
           { 
		   /*      if (!(eType & JS_EVENT_INIT))
					JoyEnabled = 1;  */
		        mask = 1 << eNumber;
                       if (eValue) Buttons |= mask;
                            else   Buttons &= ~mask;
                   }
				
	
	      case JS_EVENT_AXIS:
		/*	 if (JoyEnabled) */
	 	       { if (eNumber == 0) JoyX = eValue;
		         else
                     if (eNumber == 1) JoyY = eValue;
		       }        
	   }
       }
    
       if (!new_status) return 0;
    }
    else
#endif		/* def NEW_JOYSTCK */
    {   struct JS_DATA_TYPE js;
	int NewButtons;
	
        if (read (JoyFd, &js, JS_RETURN) != JS_RETURN) return 4;

        JoyX = js.x;
        JoyY = js.y;
 
        NewButtons = js.buttons;
	Buttons = NewButtons & ((NewButtons ^ OldButtons) | JoyButtonMask);
	OldButtons = NewButtons;

     }

/*
     mask = 2;
     while (mask)
     {  // If button pressed, ignore it if locked
        //           or set locked flag if unlocked
       if (Buttons & mask)
       {  if (JoyButtonLocked & mask) Buttons &= ~mask;
                      else JoyButtonLocked |= mask;
       }
         // If button not checked - clear locked mask
      else
            JoyButtonLocked &= ~mask;

         mask >>= 1;
      }
*/

			
     return 0;
#endif			/* def JOYSTICK_DRIVER */	
#endif			/* def JOYSTICK_SUPPORTED */
     return 4;

}

int joySetParams(void)
{
    
#ifdef JOYSTICK_SUPPORTED
#ifdef JOYSTICK_XINPUT
	if (xdev)
	{
  	    if (joyGetStatus()) return 4;

	    JoyX1 = JoyX - 5;
	    JoyX2 = JoyX + 6;
	    JoyY1 = JoyY - 5;
	    JoyY2 = JoyY + 6;
        }	    
	else
#endif	/* def JOYSTICK_XINPUT */
#ifdef NEW_JOYSTICK
	if (JoyNewDriver)
	{ 
  	    joyGetStatus();

            JoyX2 = JoyX + (long)(NEW_JOY_MAX-JoyX)*NEW_JOYSTICK_THRESHOLD_X/NEW_JOY_MAX;
            JoyX1 = JoyX - (long)(NEW_JOY_MAX+JoyX)*NEW_JOYSTICK_THRESHOLD_X/NEW_JOY_MAX;
	    JoyY2 = JoyY + (long)(NEW_JOY_MAX-JoyY)*NEW_JOYSTICK_THRESHOLD_Y/NEW_JOY_MAX;
	    JoyY1 = JoyY - (long)(NEW_JOY_MAX+JoyY)*NEW_JOYSTICK_THRESHOLD_Y/NEW_JOY_MAX;
	}
	else
#endif
	{	

  	  if (joyGetStatus()) return 4;

	  JoyX1 = JoyX >> 1;
	  JoyY1 = JoyY >> 1;

	  if (JoyX1 == 0 && JoyY1 == 0) return 4;

	  JoyX2 = JoyX1 * 3;
	  JoyY2 = JoyY1 * 3;
	}
	return 0;
#else
	return 4;
#endif
}


static int getJoy(void)
{
#ifdef JOYSTICK_SUPPORTED

   if (joyGetStatus() == 0)
   {
        if (Buttons)
        {  JoyEnabled = 1;
           return (JOY_START | Buttons);	
        }
		         
/*        printf ("JoyX = %d, JoyX1 = %d, JoyX2 = %d\n", 
	        JoyX, JoyX1, JoyX2);  */
		
      /*  No joystick button is pressed */
	if (JoyY > JoyY2)
	{ if (JoyX >= JoyX1 && JoyX <= JoyX2)
                     return JOY_DOWN;
	}
	else
	if (JoyY < JoyY1)
	{ if (JoyX >= JoyX1 && JoyX <= JoyX2)
                     return JOY_UP;
	}
	else
	if (JoyX > JoyX2) return JOY_RIGHT;
	else
	if (JoyX < JoyX1) return JOY_LEFT;
  }
#endif

  return 0;
}


 /*  Get keyboard or joystick command */
int getKeyJoy(void)
{

   int ret = getJoy();

   if ( !JoyEnabled || !ret) ret = kbdGet();

   return ret;
}


static int IndOfInt(int where[], int what)
{  int i, k;

   for (i=0; (k=where[i])!= '\0'; i++)
       if (k==what) return i;

  return -1;
}


int checkKeys(int *keys)
{  int ind;
 

#ifdef JOYSTICK_SUPPORTED
  if (joyGetStatus() == 0)
  {

    if (Buttons & 0x1)
    {  if ((ind = IndOfInt(keys, JOY_BTN1)) >= 0)
                         return ind;
    }

    if (Buttons & 0x2)
    {  if ((ind = IndOfInt(keys, JOY_BTN2)) >= 0)
                         return ind;
    }

  }
#endif

  while ((ind = kbdGet()) != 0)
  {   if (ind < 0x100) ind = toupper(ind);
      if ((ind = IndOfInt(keys, ind)) >= 0)
               return ind;
  }

  return -1;
}

int IsJoystickSupported(void)
{  
#ifdef JOYSTICK_XINPUT
    if (xdev) return 2;
    else
#endif    
    return (JoyFd < 0) ? 0 : 1;
}

